==> thx_cron-0.3.5-py3-none-any.whl: thx/cron/models.py <==
# -*- coding: utf-8 -*-
from functools import reduce

import logging
import operator

import django
from django.conf import settings
from django.core import validators
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import ExpressionWrapper, F, Q, Case, When
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

if django.VERSION < (3, 1):
    from django.contrib.postgres.fields import JSONField
else:
    from django.db.models import JSONField

from thx.core.models import UserModel, DateModel, OrderedModel, CleanModel

from . import utils
from .validators import validate_year

logger = logging.getLogger(__name__)


class Job(CleanModel, UserModel, DateModel, OrderedModel):
    status = models.IntegerField(
        choices=settings.CRON_JOB_STATUS_CHOICES,
        default=0,
        verbose_name=_('status'),
    )
    year = models.PositiveIntegerField(
        blank=True, null=True,
        validators=[validate_year],
        verbose_name=_('year'),
    )
    month = models.PositiveIntegerField(
        blank=True, null=True,
        validators=[validators.MinValueValidator(1), validators.MaxValueValidator(12)],
        verbose_name=_('month'),
    )
    day = models.PositiveIntegerField(
        blank=True, null=True,
        verbose_name=_('day'),
    )
    hour = models.PositiveIntegerField(
        blank=True, null=True,
        validators=[validators.MinValueValidator(0), validators.MaxValueValidator(23)],
        verbose_name=_('hour'),
    )
    minute = models.PositiveIntegerField(
        blank=True, null=True,
        validators=[validators.MinValueValidator(0), validators.MaxValueValidator(59)],
        verbose_name=_('minute'),
    )
    weekday = models.PositiveIntegerField(
        blank=True, null=True,
        validators=[validators.MinValueValidator(0), validators.MaxValueValidator(6)],
        verbose_name=_('weekday')
    )
    name = models.CharField(
        max_length=255,
        verbose_name=_('name'),
    )
    command = models.CharField(
        choices=utils.get_command(),
        max_length=255,
        verbose_name=_('command'),
    )
    parameters = JSONField(
        blank=True,
        default=dict,
        verbose_name=_('parameters'),
    )
    is_una_tantum = models.BooleanField(
        default=False,
        verbose_name=_('is una tantum'),
    )
    priority = models.IntegerField(
        choices=settings.CRON_JOB_PRIORITY_CHOICES,
        default=4,  # normal
        verbose_name=_("priority")
    )
    # Alert
    max_notifications = models.IntegerField(
        default=5,
        verbose_name=_("max notifications")
    )
    sent_notifications_count = models.IntegerField(
        default=0,
        verbose_name=_("sent notifications count")
    )
    max_job_duration = models.DurationField(
        blank=True, null=True,
        verbose_name=_("max job duration")
    )

    class Meta:
        verbose_name = _('job')
        verbose_name_plural = _('jobs')
        ordering = ('priority', 'ordering',)
        permissions = (
            ("list_job", "Can list job"),
            ("detail_job", "Can detail job"),
            ("disable_job", "Can disable job"),
        )

    def __str__(self):
        return self.name

    def clean(self):
        self.check_date()
        self.check_parameters()
        self.check_max_job_duration()
        super().clean()

    def check_date(self):
        if self.year and self.month and self.day:
            try:
                timezone.datetime(self.year, self.month, self.day)
            except ValueError:
                raise ValidationError(_("insert valid year, month and day"))

        elif self.year is None and self.month and self.day:
            # Check on next 5 years...
            this_year = timezone.now().year
            try:
                for year in [this_year + i for i in range(0, 6)]:
                    timezone.datetime(year, self.month, self.day)
            except ValueError:
                raise ValidationError(_("insert valid month and day"))

    def check_max_job_duration(self):
        if self.max_job_duration and self.max_job_duration < timezone.timedelta(minutes=2):
            raise ValidationError({'max_job_duration': _("set at least 2 minutes")})

    def get_max_duration(self):
        return self.max_job_duration or settings.CRON_JOB_MAX_JOB_DURATION

    def check_parameters(self):
        if self.parameters is None:
            raise ValidationError(_("value must be valid JSON"))

    @property
    def is_ripetitive(self):
        if self.is_una_tantum:
            return False
        return not all(v is not None for v in [self.year, self.month, self.day, self.hour, self.minute])

    @property
    def last_execution_status(self):
        last_job_execution = self.job_elaborations.all().order_by('date_create').last()
        if last_job_execution:
            return last_job_execution.is_valid

    def set_running(self, info=None):
        running = -1
        self.status = running
        self.save()

    def set_completed(self, info=None):
        completed = 1
        self.status = completed
        self.save()

    def set_to_run(self, info=None):
        to_run = 0
        self.status = to_run
        self.save()

    def set_error(self, info=None):
        completed = 1
        self.status = completed
        self.save()

    def run_command(self):
        command = import_string(self.command)
        result = command(instance=self, **self.parameters).do()
        return result

    def send_alert(self):
        """
        Send alert to the configured accounts.
        Use the configured backend. Default: CronAlertEmailBackend
        """
        result = import_string(settings.CRON_ALERT_BACKEND)(job=self).send_alert()
        if result:
            self.sent_notifications_count += 1
            self.save(update_fields=['sent_notifications_count'])
        return result

    @classmethod
    def selected_jobs(cls, datetime=None, status=0):
        if datetime is None:
            datetime = timezone.now()
        qs = cls.objects.filter(status=status)
        query1 = Q(is_una_tantum=True)
        query2 = Q(year=datetime.year) | Q(year__isnull=True)
        query2 &= Q(month=datetime.month) | Q(month__isnull=True)
        query2 &= Q(day=datetime.day) | Q(day__isnull=True)
        query2 &= Q(hour=datetime.hour) | Q(hour__isnull=True)
        query2 &= Q(minute=datetime.minute) | Q(minute__isnull=True)
        query2 &= Q(weekday=datetime.weekday()) | Q(weekday__isnull=True)
        qs = qs.filter(query1 | query2)

        # TODO test
        query_list = [
            Q(year=datetime.year) | Q(year__isnull=True),
            Q(month=datetime.month) | Q(month__isnull=True),
            Q(day=datetime.day) | Q(day__isnull=True),
            Q(hour=datetime.hour) | Q(hour__isnull=True),
            Q(minute=datetime.minute) | Q(minute__isnull=True),
            Q(weekday=datetime.weekday()) | Q(weekday__isnull=True)
        ]
        qs = qs.filter(query1 | reduce(operator.and_, query_list))

        return qs

    @classmethod
    def complete_old_jobs(cls, datetime=None):
        if datetime is None:
            datetime = timezone.now()
        to_run = 0
        completed = 1
        count = 0
        qs = cls.objects.filter(status=to_run, is_una_tantum=False)
        count += qs.filter(year__lt=datetime.year).update(status=completed)
        count += qs.filter(year=datetime.year, month__lt=datetime.month).update(status=completed)
        count += qs.filter(year=datetime.year, month=datetime.month, day__lt=datetime.day).update(status=completed)
        count += qs.filter(
            year=datetime.year, month=datetime.month, day=datetime.day,
            hour__lt=datetime.hour
        ).update(status=completed)
        count += qs.filter(
            year=datetime.year, month=datetime.month, day=datetime.day,
            hour=datetime.hour, minute__lt=datetime.minute
        ).update(status=completed)
        return count

    @classmethod
    def verify_old_jobs_and_send_alert_if_necessary(cls, datetime=None):
        if datetime is None:
            datetime = timezone.now()

        running = -1
        default = settings.CRON_JOB_MAX_JOB_DURATION
        extra_filter = Q(max_job_duration__isnull=False) if default is None else Q()

        qs = cls.objects.annotate(
            _date_check=ExpressionWrapper(
                expression=F('date_last_modify') + Case(
                    When(Q(max_job_duration__isnull=False), then=F('max_job_duration')),
                    default=default or timezone.timedelta(days=365)
                ),
                output_field=models.DateTimeField()
            ),
            _remaining_notifications=F('max_notifications') - F('sent_notifications_count'),
        ).filter(
            status=running,  # RUNNING
            # Annotated filters
            _date_check__lt=datetime,
            _remaining_notifications__gt=0
        ).filter(extra_filter)

        errors = []
        for job in qs:
            try:
                job.send_alert()
            except Exception as e:
                logger.error(e)
                errors.append(e)

        if errors:
            raise errors[0]


class JobLog(DateModel, CleanModel):
    job = models.ForeignKey(
        "Job",
        on_delete=models.CASCADE,
        related_name='job_elaborations',
        verbose_name=_('job'),
    )
    elaboration_time = models.DurationField(
        null=True, blank=True,
        verbose_name=_("elaboration time"),
    )
    elaboration_results = JSONField(
        verbose_name=_('elaboration results'),
    )
    is_valid = models.BooleanField(
        blank=True, null=True,
        verbose_name=_('is valid'),
    )
    extra_info = JSONField(
        blank=True, null=True,
        verbose_name=_('extra info')
    )
    log_file_content = models.TextField(
        blank=True, null=True,
        verbose_name=_('log file content')
    )

    class Meta:
        verbose_name = _('job log')
        verbose_name_plural = _('job logs')
        ordering = ('-date_create',)
        permissions = (
            ("list_joblog", "Can list job log"),
            ("detail_joblog", "Can detail job log"),
            ("disable_joblog", "Can disable job log"),
        )

    def __str__(self):
        return str(self.date_create)

    def clean(self):
        pass

==> thx_cron-0.4.4-py3-none-any.whl: thx/cron/models.py <==
# -*- coding: utf-8 -*-
from functools import reduce

import logging
import operator

import django
from django.conf import settings
from django.core import validators
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import ExpressionWrapper, F, Q, Case, When
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

if django.VERSION < (3, 1):
    from django.contrib.postgres.fields import JSONField
else:
    from django.db.models import JSONField

from thx.core.models import UserModel, DateModel, OrderedModel, CleanModel

from . import utils
from .validators import validate_year

logger = logging.getLogger(__name__)


class Job(CleanModel, UserModel, DateModel, OrderedModel):
    status = models.IntegerField(
        choices=settings.CRON_JOB_STATUS_CHOICES,
        default=0,
        verbose_name=_('status'),
    )
    year = models.PositiveIntegerField(
        blank=True, null=True,
        validators=[validate_year],
        verbose_name=_('year'),
    )
    month = models.PositiveIntegerField(
        blank=True, null=True,
        validators=[validators.MinValueValidator(1), validators.MaxValueValidator(12)],
        verbose_name=_('month'),
    )
    day = models.PositiveIntegerField(
        blank=True, null=True,
        verbose_name=_('day'),
    )
    hour = models.PositiveIntegerField(
        blank=True, null=True,
        validators=[validators.MinValueValidator(0), validators.MaxValueValidator(23)],
        verbose_name=_('hour'),
    )
    minute = models.PositiveIntegerField(
        blank=True, null=True,
        validators=[validators.MinValueValidator(0), validators.MaxValueValidator(59)],
        verbose_name=_('minute'),
    )
    weekday = models.PositiveIntegerField(
        blank=True, null=True,
        validators=[validators.MinValueValidator(0), validators.MaxValueValidator(6)],
        verbose_name=_('weekday')
    )
    name = models.CharField(
        max_length=255,
        verbose_name=_('name'),
    )
    command = models.CharField(
        choices=utils.get_command(),
        max_length=255,
        verbose_name=_('command'),
    )
    parameters = JSONField(
        blank=True,
        default=dict,
        verbose_name=_('parameters'),
    )
    is_una_tantum = models.BooleanField(
        default=False,
        verbose_name=_('is una tantum'),
    )
    priority = models.IntegerField(
        choices=settings.CRON_JOB_PRIORITY_CHOICES,
        default=4,  # normal
        verbose_name=_("priority")
    )
    # Alert
    max_notifications = models.IntegerField(
        default=5,
        verbose_name=_("max notifications")
    )
    sent_notifications_count = models.IntegerField(
        default=0,
        verbose_name=_("sent notifications count")
    )
    max_job_duration = models.DurationField(
        blank=True, null=True,
        verbose_name=_("max job duration")
    )

    class Meta:
        verbose_name = _('job')
        verbose_name_plural = _('jobs')
        ordering = ('priority', 'ordering',)
        permissions = (
            ("list_job", "Can list job"),
            ("detail_job", "Can detail job"),
            ("disable_job", "Can disable job"),
        )

    def __str__(self):
        return self.name

    def clean(self):
        self.check_date()
        self.check_parameters()
        self.check_max_job_duration()
        super().clean()

    def check_date(self):
        if self.year and self.month and self.day:
            try:
                timezone.datetime(self.year, self.month, self.day)
            except ValueError:
                raise ValidationError(_("insert valid year, month and day"))

        elif self.year is None and self.month and self.day:
            # Check on next 5 years...
            this_year = timezone.now().year
            try:
                for year in [this_year + i for i in range(0, 6)]:
                    timezone.datetime(year, self.month, self.day)
            except ValueError:
                raise ValidationError(_("insert valid month and day"))

    def check_max_job_duration(self):
        if self.max_job_duration and self.max_job_duration < timezone.timedelta(minutes=2):
            raise ValidationError({'max_job_duration': _("set at least 2 minutes")})

    def get_max_duration(self):
        return self.max_job_duration or settings.CRON_JOB_MAX_JOB_DURATION

    def check_parameters(self):
        if self.parameters is None:
            raise ValidationError(_("value must be valid JSON"))

    @property
    def is_ripetitive(self):
        if self.is_una_tantum:
            return False
        return not all(v is not None for v in [self.year, self.month, self.day, self.hour, self.minute])

    @property
    def last_execution_status(self):
        last_job_execution = self.job_elaborations.all().order_by('date_create').last()
        if last_job_execution:
            return last_job_execution.is_valid

    def set_running(self, info=None):
        running = -1
        self.status = running
        self.save()

    def set_completed(self, info=None):
        completed = 1
        self.status = completed
        self.save()

    def set_to_run(self, info=None):
        to_run = 0
        self.status = to_run
        self.save()

    def set_error(self, info: dict):
        completed = 1
        self.status = completed
        self.save()
        try:
            JobLog.objects.create(
                job=self,
                elaboration_time=None,
                elaboration_results=info,
                extra_info=None,
                is_valid=False,
                log_file_content=None
            )
        except Exception as e:
            logger.error(e)

    def run_command(self):
        command = import_string(self.command)
        result = command(instance=self, **self.parameters).do()
        return result

    def send_alert(self):
        """
        Send alert to the configured accounts.
        Use the configured backend. Default: CronAlertEmailBackend
        """
        result = import_string(settings.CRON_ALERT_BACKEND)(job=self).send_alert()
        if result:
            self.sent_notifications_count += 1
            self.save(update_fields=['sent_notifications_count'])
        return result

    @classmethod
    def selected_jobs(cls, datetime=None, status=0):
        if datetime is None:
            datetime = timezone.now()
        qs = cls.objects.filter(status=status)
        query1 = Q(is_una_tantum=True)
        query2 = Q(year=datetime.year) | Q(year__isnull=True)
        query2 &= Q(month=datetime.month) | Q(month__isnull=True)
        query2 &= Q(day=datetime.day) | Q(day__isnull=True)
        query2 &= Q(hour=datetime.hour) | Q(hour__isnull=True)
        query2 &= Q(minute=datetime.minute) | Q(minute__isnull=True)
        query2 &= Q(weekday=datetime.weekday()) | Q(weekday__isnull=True)
        qs = qs.filter(query1 | query2)

        # TODO test
        query_list = [
            Q(year=datetime.year) | Q(year__isnull=True),
            Q(month=datetime.month) | Q(month__isnull=True),
            Q(day=datetime.day) | Q(day__isnull=True),
            Q(hour=datetime.hour) | Q(hour__isnull=True),
            Q(minute=datetime.minute) | Q(minute__isnull=True),
            Q(weekday=datetime.weekday()) | Q(weekday__isnull=True)
        ]
        qs = qs.filter(query1 | reduce(operator.and_, query_list))

        return qs

    @classmethod
    def complete_old_jobs(cls, datetime=None):
        if datetime is None:
            datetime = timezone.now()
        to_run = 0
        completed = 1
        count = 0
        qs = cls.objects.filter(status=to_run, is_una_tantum=False)
        count += qs.filter(year__lt=datetime.year).update(status=completed)
        count += qs.filter(year=datetime.year, month__lt=datetime.month).update(status=completed)
        count += qs.filter(year=datetime.year, month=datetime.month, day__lt=datetime.day).update(status=completed)
        count += qs.filter(
            year=datetime.year, month=datetime.month, day=datetime.day,
            hour__lt=datetime.hour
        ).update(status=completed)
        count += qs.filter(
            year=datetime.year, month=datetime.month, day=datetime.day,
            hour=datetime.hour, minute__lt=datetime.minute
        ).update(status=completed)
        return count

    @classmethod
    def verify_old_jobs_and_send_alert_if_necessary(cls, datetime=None):
        if datetime is None:
            datetime = timezone.now()

        running = -1
        default = settings.CRON_JOB_MAX_JOB_DURATION
        extra_filter = Q(max_job_duration__isnull=False) if default is None else Q()

        qs = cls.objects.annotate(
            _date_check=ExpressionWrapper(
                expression=F('date_last_modify') + Case(
                    When(Q(max_job_duration__isnull=False), then=F('max_job_duration')),
                    default=default or timezone.timedelta(days=365)
                ),
                output_field=models.DateTimeField()
            ),
            _remaining_notifications=F('max_notifications') - F('sent_notifications_count'),
        ).filter(
            status=running,  # RUNNING
            # Annotated filters
            _date_check__lt=datetime,
            _remaining_notifications__gt=0
        ).filter(extra_filter)

        errors = []
        for job in qs:
            try:
                job.send_alert()
            except Exception as e:
                logger.error(e)
                errors.append(e)

        if errors:
            raise errors[0]


class JobLog(DateModel, CleanModel):
    job = models.ForeignKey(
        "Job",
        on_delete=models.CASCADE,
        related_name='job_elaborations',
        verbose_name=_('job'),
    )
    elaboration_time = models.DurationField(
        null=True, blank=True,
        verbose_name=_("elaboration time"),
    )
    elaboration_results = JSONField(
        verbose_name=_('elaboration results'),
    )
    is_valid = models.BooleanField(
        blank=True, null=True,
        verbose_name=_('is valid'),
    )
    extra_info = JSONField(
        blank=True, null=True,
        verbose_name=_('extra info')
    )
    log_file_content = models.TextField(
        blank=True, null=True,
        verbose_name=_('log file content')
    )

    class Meta:
        verbose_name = _('job log')
        verbose_name_plural = _('job logs')
        ordering = ('-date_create',)
        permissions = (
            ("list_joblog", "Can list job log"),
            ("detail_joblog", "Can detail job log"),
            ("disable_joblog", "Can disable job log"),
        )

    def __str__(self):
        return str(self.date_create)

    def clean(self):
        pass
